from pprint import pformat


class Model(object):
    """
    Implements a generic object.
    """

    def __init__(self, data, api):
        self.temp_id = ""
        self.data = data
        self.api = api

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def __repr__(self):
        formatted_dict = pformat(dict(self.data))
        classname = self.__class__.__name__
        return "%s(%s)" % (classname, formatted_dict)

    def __contains__(self, value):
        return value in self.data


class Collaborator(Model):
    """
    Implements a collaborator.
    """

    def delete(self, project_id):
        """
        Deletes a collaborator from a shared project.
        """
        self.api.collaborators.delete(project_id, self["email"])


class CollaboratorState(Model):
    """
    Implements a collaborator state.
    """

    pass


class Filter(Model):
    """
    Implements a filter.
    """

    def update(self, **kwargs):
        """
        Updates filter.
        """
        self.api.filters.update(self["id"], **kwargs)
        self.data.update(kwargs)

    def delete(self):
        """
        Deletes filter.
        """
        self.api.filters.delete(self["id"])
        self.data["is_deleted"] = 1


class Item(Model):
    """
    Implements an item.
    """

    def update(self, **kwargs):
        """
        Updates item.
        """
        self.api.items.update(self["id"], **kwargs)
        self.data.update(kwargs)

    def delete(self):
        """
        Deletes item.
        """
        self.api.items.delete(self["id"])
        self.data["is_deleted"] = 1

    def move(self, **kwargs):
        """
        Moves item to another parent, project, or section.
        """
        if "parent_id" in kwargs:
            self.api.items.move(self["id"], parent_id=kwargs.get("parent_id"))
            self.data["parent_id"] = kwargs.get("parent_id")
        elif "project_id" in kwargs:
            self.api.items.move(self["id"], project_id=kwargs.get("project_id"))
            self.data["project_id"] = kwargs.get("project_id")
        elif "section_id" in kwargs:
            self.api.items.move(self["id"], section_id=kwargs.get("section_id"))
            self.data["section_id"] = kwargs.get("section_id")
        else:
            raise TypeError("move() takes one of parent_id, project_id, or section_id arguments")

    def reorder(self, child_order):
        """
        Reorder item.
        """
        self.api.items.reorder([{"id": self["id"], "child_order": child_order}])
        self.data["child_order"] = child_order

    def close(self):
        """
        Marks item as closed
        """
        self.api.items.close(self["id"])

    def complete(self, date_completed=None):
        """
        Marks item as completed.
        """
        self.api.items.complete(self["id"], date_completed=date_completed)
        self.data["checked"] = 1

    def uncomplete(self):
        """
        Marks item as uncompleted.
        """
        self.api.items.uncomplete(self["id"])
        self.data["checked"] = 0

    def archive(self):
        """
        Marks item as archived.
        """
        self.api.items.archive(self["id"])
        self.data["in_history"] = 1

    def unarchive(self):
        """
        Marks item as unarchived.
        """
        self.api.items.unarchive(self["id"])
        self.data["in_history"] = 0

    def update_date_complete(self, due=None):
        """
        Completes a recurring task.
        """
        self.api.items.update_date_complete(self["id"], due=due)
        if due:
            self.data["due"] = due


class Label(Model):
    """
    Implements a label.
    """

    def update(self, **kwargs):
        """
        Updates label.
        """
        self.api.labels.update(self["id"], **kwargs)
        self.data.update(kwargs)

    def delete(self):
        """
        Deletes label.
        """
        self.api.labels.delete(self["id"])
        self.data["is_deleted"] = 1


class LiveNotification(Model):
    """
    Implements a live notification.
    """

    pass


class GenericNote(Model):
    """
    Implements a note.
    """

    #: has to be defined in subclasses
    local_manager = None

    def update(self, **kwargs):
        """
        Updates note.
        """
        self.local_manager.update(self["id"], **kwargs)
        self.data.update(kwargs)

    def delete(self):
        """
        Deletes note.
        """
        self.local_manager.delete(self["id"])
        self.data["is_deleted"] = 1


class Note(GenericNote):
    """
    Implement an item note.
    """

    def __init__(self, data, api):
        GenericNote.__init__(self, data, api)
        self.local_manager = self.api.notes


class ProjectNote(GenericNote):
    """
    Implement a project note.
    """

    def __init__(self, data, api):
        GenericNote.__init__(self, data, api)
        self.local_manager = self.api.project_notes


class Project(Model):
    """
    Implements a project.
    """

    def update(self, **kwargs):
        """
        Updates project.
        """
        self.api.projects.update(self["id"], **kwargs)
        self.data.update(kwargs)

    def delete(self):
        """
        Deletes project.
        """
        self.api.projects.delete(self["id"])
        self.data["is_deleted"] = 1

    def archive(self):
        """
        Marks project as archived.
        """
        self.api.projects.archive(self["id"])
        self.data["is_archived"] = 1

    def unarchive(self):
        """
        Marks project as unarchived.
        """
        self.api.projects.unarchive(self["id"])
        self.data["is_archived"] = 0

    def move(self, parent_id):
        """
        Moves project to another parent.
        """
        self.api.projects.move(self["id"], parent_id)

    def reorder(self, child_order):
        """
        Reorder project.
        """
        self.api.projects.reorder([{"id": self["id"], "child_order": child_order}])
        self.data["child_order"] = child_order

    def share(self, email):
        """
        Shares projects with a user.
        """
        self.api.projects.share(self["id"], email)

    def take_ownership(self):
        """
        Takes ownership of a shared project.
        """
        self.api.projects.take_ownership(self["id"])


class Reminder(Model):
    """
    Implements a reminder.
    """

    def update(self, **kwargs):
        """
        Updates reminder.
        """
        self.api.reminders.update(self["id"], **kwargs)
        self.data.update(kwargs)

    def delete(self):
        """
        Deletes reminder.
        """
        self.api.reminders.delete(self["id"])
        self.data["is_deleted"] = 1


class Section(Model):
    """
    Implements a section.
    """

    def update(self, **kwargs):
        """
        Updates section.
        """
        self.api.sections.update(self["id"], **kwargs)
        self.data.update(kwargs)

    def delete(self):
        """
        Deletes section.
        """
        self.api.sections.delete(self["id"])
        self.data["is_deleted"] = 1

    def move(self, project_id):
        """
        Moves section to another project.
        """
        self.api.sections.move(self["id"], project_id=project_id)
        self.data["project_id"] = project_id

    def reorder(self, section_order):
        """
        Reorder section.
        """
        self.api.sections.reorder([{"id": self["id"], "section_order": section_order}])
        self.data["section_order"] = section_order

    def archive(self, date_archived=None):
        """
        Marks section as archived.
        """
        self.api.sections.archive(self["id"], date_archived=date_archived)
        self.data["is_archived"] = 1

    def unarchive(self):
        """
        Marks section as unarchived.
        """
        self.api.sections.unarchive(self["id"])
        self.data["is_archived"] = 0



# -*- coding: utf-8 -*-
from .generic import AllMixin, Manager, SyncMixin


class LocationsManager(Manager, AllMixin, SyncMixin):

    state_name = "locations"
    object_type = None  # there is no local state associated

    def clear(self):
        """
        Clears the locations.
        """
        cmd = {
            "type": "clear_locations",
            "uuid": self.api.generate_uuid(),
            "args": {},
        }
        self.queue.append(cmd)



# -*- coding: utf-8 -*-
from .. import models
from .generic import AllMixin, GetByIdMixin, Manager, SyncMixin


class ItemsManager(Manager, AllMixin, GetByIdMixin, SyncMixin):

    state_name = "items"
    object_type = "item"

    def add(self, content, **kwargs):
        """
        Creates a local item object.
        """
        project_id = kwargs.get("project_id")
        if not project_id:
            project_id = self.state["user"]["inbox_project"]
        obj = models.Item({"content": content, "project_id": project_id}, self.api)
        obj.temp_id = obj["id"] = self.api.generate_uuid()
        obj.data.update(kwargs)
        self.state[self.state_name].append(obj)
        cmd = {
            "type": "item_add",
            "temp_id": obj.temp_id,
            "uuid": self.api.generate_uuid(),
            "args": {key: obj.data[key] for key in obj.data if key != "id"},
        }
        self.queue.append(cmd)
        return obj

    def update(self, item_id, **kwargs):
        """
        Updates an item remotely.
        """
        args = {"id": item_id}
        args.update(kwargs)
        cmd = {
            "type": "item_update",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def delete(self, item_id):
        """
        Deletes an item remotely.
        """
        cmd = {
            "type": "item_delete",
            "uuid": self.api.generate_uuid(),
            "args": {"id": item_id},
        }
        self.queue.append(cmd)

    def move(self, item_id, **kwargs):
        """
        Moves item to another parent, project, or section remotely.
        """
        args = {
            "id": item_id,
        }
        if "parent_id" in kwargs:
            args["parent_id"] = kwargs.get("parent_id")
        elif "project_id" in kwargs:
            args["project_id"] = kwargs.get("project_id")
        elif "section_id" in kwargs:
            args["section_id"] = kwargs.get("section_id")
        else:
            raise TypeError("move() takes one of parent_id, project_id, or section_id arguments")
        cmd = {"type": "item_move", "uuid": self.api.generate_uuid(), "args": args}
        self.queue.append(cmd)

    def close(self, item_id):
        """
        Marks item as done
        """
        cmd = {
            "type": "item_close",
            "uuid": self.api.generate_uuid(),
            "args": {"id": item_id},
        }
        self.queue.append(cmd)

    def complete(self, item_id, date_completed=None, force_history=None):
        """
        Marks item as completed remotely.
        """
        args = {
            "id": item_id,
        }
        if date_completed is not None:
            args["date_completed"] = date_completed
        if force_history is not None:
            args["force_history"] = force_history
        cmd = {
            "type": "item_complete",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def uncomplete(self, item_id):
        """
        Marks item as uncompleted remotely.
        """
        cmd = {
            "type": "item_uncomplete",
            "uuid": self.api.generate_uuid(),
            "args": {"id": item_id},
        }
        self.queue.append(cmd)

    def archive(self, item_id):
        """
        Marks item as archived remotely.
        """
        cmd = {
            "type": "item_archive",
            "uuid": self.api.generate_uuid(),
            "args": {"id": item_id},
        }
        self.queue.append(cmd)

    def unarchive(self, item_id):
        """
        Marks item as unarchived remotely.
        """
        cmd = {
            "type": "item_unarchive",
            "uuid": self.api.generate_uuid(),
            "args": {"id": item_id},
        }
        self.queue.append(cmd)

    def update_date_complete(self, item_id, due=None):
        """
        Completes a recurring task remotely.
        """
        args = {
            "id": item_id,
        }
        if due:
            args["due"] = due
        cmd = {
            "type": "item_update_date_complete",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def reorder(self, items):
        """
        Updates the child_order of the specified items.
        """
        cmd = {
            "type": "item_reorder",
            "uuid": self.api.generate_uuid(),
            "args": {"items": items},
        }
        self.queue.append(cmd)

    def update_day_orders(self, ids_to_orders):
        """
        Updates in the local state the day orders of multiple items remotely.
        """
        cmd = {
            "type": "item_update_day_orders",
            "uuid": self.api.generate_uuid(),
            "args": {"ids_to_orders": ids_to_orders},
        }
        self.queue.append(cmd)

    def get_completed(self, project_id, **kwargs):
        """
        Returns a project's completed items.
        """
        params = {"token": self.token, "project_id": project_id}
        params.update(kwargs)
        return self.api._get("items/get_completed", params=params)

    def get(self, item_id):
        """
        Gets an existing item.
        """
        params = {"token": self.token, "item_id": item_id}
        obj = self.api._get("items/get", params=params)
        if obj and "error" in obj:
            return None

        data = {"projects": [], "items": [], "notes": []}
        if obj.get("project"):
            data["projects"].append(obj.get("project"))
        if obj.get("item"):
            data["items"].append(obj.get("item"))
        if obj.get("notes"):
            data["notes"] += obj.get("notes")
        self.api._update_state(data)
        return obj



# -*- coding: utf-8 -*-
class Manager(object):

    # should be re-defined in a subclass
    state_name = None
    object_type = None

    def __init__(self, api):
        self.api = api

    # shortcuts
    @property
    def state(self):
        return self.api.state

    @property
    def queue(self):
        return self.api.queue

    @property
    def token(self):
        return self.api.token


class AllMixin(object):
    def all(self, filt=None):
        return list(filter(filt, self.state[self.state_name]))


class GetByIdMixin(object):
    def get_by_id(self, obj_id, only_local=False):
        """
        Finds and returns the object based on its id.
        """
        for obj in self.state[self.state_name]:
            if obj["id"] == obj_id or obj.temp_id == str(obj_id):
                return obj

        if not only_local and self.object_type is not None:
            getter = getattr(eval("self.api.%ss" % self.object_type), "get")
            data = getter(obj_id)

            # retrieves from state, otherwise we return the raw data
            for obj in self.state[self.state_name]:
                if obj["id"] == obj_id or obj.temp_id == str(obj_id):
                    return obj

            return data

        return None


class SyncMixin(object):
    """
    Syncs this specific type of objects.
    """

    def sync(self):
        return self.api.sync()



# -*- coding: utf-8 -*-
from .generic import Manager


class BackupsManager(Manager):
    def get(self):
        """
        Get backups.
        """
        params = {"token": self.token}
        return self.api._get("backups/get", params=params)



# -*- coding: utf-8 -*-
from .generic import Manager


class QuickManager(Manager):
    def add(self, text, **kwargs):
        """
        Quick add task implementation.
        """
        params = {"token": self.token, "text": text}
        params.update(kwargs)
        return self.api._get("quick/add", params=params)



# -*- coding: utf-8 -*-
from .. import models
from .generic import AllMixin, GetByIdMixin, Manager, SyncMixin


class SectionsManager(Manager, AllMixin, GetByIdMixin, SyncMixin):

    state_name = "sections"
    object_type = "section"

    def add(self, name, project_id, **kwargs):
        """
        Creates a local section object.
        """
        obj = models.Section({"name": name, "project_id": project_id}, self.api)
        obj.temp_id = obj["id"] = self.api.generate_uuid()
        obj.data.update(kwargs)
        self.state[self.state_name].append(obj)
        cmd = {
            "type": "section_add",
            "temp_id": obj.temp_id,
            "uuid": self.api.generate_uuid(),
            "args": {key: obj.data[key] for key in obj.data if key != "id"},
        }
        self.queue.append(cmd)
        return obj

    def update(self, section_id, **kwargs):
        """
        Updates a section remotely.
        """
        args = {"id": section_id}
        args.update(kwargs)
        cmd = {
            "type": "section_update",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def delete(self, section_id):
        """
        Delete a section remotely.
        """
        cmd = {
            "type": "section_delete",
            "uuid": self.api.generate_uuid(),
            "args": {"id": section_id},
        }
        self.queue.append(cmd)

    def move(self, section_id, project_id):
        """
        Moves section to another project remotely.
        """
        cmd = {
            "type": "section_move",
            "uuid": self.api.generate_uuid(),
            "args": {"id": section_id, "project_id": project_id},
        }
        self.queue.append(cmd)

    def archive(self, section_id, **kwargs):
        """
        Marks section as archived remotely.
        """
        args = {
            "id": section_id,
        }
        if kwargs.get("date_archived"):
            args["date_archived"] = kwargs.get("date_archived")
        cmd = {
            "type": "section_archive",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def unarchive(self, section_id):
        """
        Marks section as unarchived remotely.
        """
        cmd = {
            "type": "section_unarchive",
            "uuid": self.api.generate_uuid(),
            "args": {"id": section_id},
        }
        self.queue.append(cmd)

    def reorder(self, sections):
        """
        Updates the section_order of the specified sections.
        """
        cmd = {
            "type": "section_reorder",
            "uuid": self.api.generate_uuid(),
            "args": {"sections": sections},
        }
        self.queue.append(cmd)

    def get(self, section_id):
        """
        Gets an existing section.
        """
        params = {"token": self.token, "section_id": section_id}
        obj = self.api._get("sections/get", params=params)
        if obj and "error" in obj:
            return None
        data = {"sections": []}
        if obj.get("section"):
            data["sections"].append(obj.get("section"))
        self.api._update_state(data)
        return obj



# -*- coding: utf-8 -*-
from .generic import Manager, SyncMixin


class InvitationsManager(Manager, SyncMixin):

    state_name = None  # there is no local state associated
    object_type = "share_invitation"

    def accept(self, invitation_id, invitation_secret):
        """
        Accepts an invitation to share a project.
        """
        cmd = {
            "type": "accept_invitation",
            "uuid": self.api.generate_uuid(),
            "args": {
                "invitation_id": invitation_id,
                "invitation_secret": invitation_secret,
            },
        }
        self.queue.append(cmd)

    def reject(self, invitation_id, invitation_secret):
        """
        Rejets an invitation to share a project.
        """
        cmd = {
            "type": "reject_invitation",
            "uuid": self.api.generate_uuid(),
            "args": {
                "invitation_id": invitation_id,
                "invitation_secret": invitation_secret,
            },
        }
        self.queue.append(cmd)

    def delete(self, invitation_id):
        """
        Delete an invitation to share a project.
        """
        cmd = {
            "type": "delete_invitation",
            "uuid": self.api.generate_uuid(),
            "args": {"invitation_id": invitation_id},
        }
        self.queue.append(cmd)



# -*- coding: utf-8 -*-
from .generic import Manager


class CompletedManager(Manager):
    def get_stats(self):
        """
        Returns the user's recent productivity stats.
        """
        return self.api._get("completed/get_stats", params={"token": self.token})

    def get_all(self, **kwargs):
        """
        Returns all user's completed items.
        """
        params = {"token": self.token}
        params.update(kwargs)
        return self.api._get("completed/get_all", params=params)



# -*- coding: utf-8 -*-
from .generic import GetByIdMixin, Manager, SyncMixin


class CollaboratorsManager(Manager, GetByIdMixin, SyncMixin):

    state_name = "collaborators"
    object_type = None  # there is no object type associated

    def delete(self, project_id, email):
        """
        Deletes a collaborator from a shared project.
        """
        cmd = {
            "type": "delete_collaborator",
            "uuid": self.api.generate_uuid(),
            "args": {"project_id": project_id, "email": email},
        }
        self.queue.append(cmd)



# -*- coding: utf-8 -*-
from .generic import Manager


class ActivityManager(Manager):
    def get(self, **kwargs):
        """
        Get events from the activity log.
        """
        params = {"token": self.token}
        params.update(kwargs)
        return self.api._get("activity/get", params=params)



import io
import time

import todoist


def test_stats_get(api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    response = api.completed.get_stats()
    assert "days_items" in response
    assert "week_items" in response
    assert "karma_trend" in response
    assert "karma_last_update" in response


def test_user_update(api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()
    date_format = api.state["user"]["date_format"]
    date_format_new = 1 - date_format
    api.user.update(date_format=date_format_new)
    api.commit()
    assert date_format_new == api.state["user"]["date_format"]
    api.user.update_goals(vacation_mode=1)
    api.commit()
    api.user.update_goals(vacation_mode=0)
    api.commit()


def test_user_settings_update(api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()
    reminder_email = api.state["user_settings"]["reminder_email"]
    if reminder_email:
        reminder_email = False
    else:
        reminder_email = True
    api.user_settings.update(reminder_email=reminder_email)
    api.commit()
    assert reminder_email == api.state["user_settings"]["reminder_email"]


def test_project_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    response = api.commit()
    assert response["projects"][0]["name"] == "Project1"
    assert "Project1" in [p["name"] for p in api.state["projects"]]
    assert api.projects.get_by_id(project1["id"]) == project1

    assert api.projects.get(project1["id"])["project"]["name"] == project1["name"]

    project1.delete()
    api.commit()


def test_project_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    project1.delete()
    response = api.commit()
    assert response["projects"][0]["id"] == project1["id"]
    assert response["projects"][0]["is_deleted"] == 1
    assert "Project1" not in [p["name"] for p in api.state["projects"]]


def test_project_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    project1.update(name="UpdatedProject1")
    response = api.commit()
    assert response["projects"][0]["name"] == "UpdatedProject1"
    assert "UpdatedProject1" in [p["name"] for p in api.state["projects"]]
    assert api.projects.get_by_id(project1["id"]) == project1

    project1.delete()
    api.commit()


def test_project_archive(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    project1.archive()
    response = api.commit()
    assert response["projects"][0]["name"] == "Project1"
    assert response["projects"][0]["is_archived"] == 1
    assert "Project1" in [p["name"] for p in api.state["projects"]]
    assert 1 in [
        p["is_archived"] for p in api.state["projects"] if p["id"] == project1["id"]
    ]

    project1.delete()
    api.commit()


def test_project_unarchive(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    project1.archive()
    api.commit()

    project1.unarchive()
    response = api.commit()
    assert response["projects"][0]["name"] == "Project1"
    assert response["projects"][0]["is_archived"] == 0
    assert 0 in [
        p["is_archived"] for p in api.state["projects"] if p["id"] == project1["id"]
    ]

    project1.delete()
    api.commit()


def test_project_move_to_parent(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    project2 = api.projects.add("Project2")
    api.commit()

    project2.move(project1["id"])
    response = api.commit()
    assert response["projects"][0]["name"] == "Project2"
    assert response["projects"][0]["parent_id"] == project1["id"]
    assert project1["id"] in [
        i["parent_id"] for i in api.state["projects"] if i["id"] == project2["id"]
    ]

    project2.delete()
    api.commit()
    project1.delete()
    api.commit()


def test_project_reorder(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    project2 = api.projects.add("Project2")
    api.commit()

    api.projects.reorder(
        projects=[
            {"id": project1["id"], "child_order": 2},
            {"id": project2["id"], "child_order": 1},
        ]
    )
    response = api.commit()
    for project in response["projects"]:
        if project["id"] == project1["id"]:
            assert project["child_order"] == 2
        if project["id"] == project2["id"]:
            assert project["child_order"] == 1
    assert 2 in [
        p["child_order"] for p in api.state["projects"] if p["id"] == project1["id"]
    ]
    assert 1 in [
        p["child_order"] for p in api.state["projects"] if p["id"] == project2["id"]
    ]

    project1.delete()
    api.commit()
    project2.delete()
    api.commit()


def test_item_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    response = api.add_item("Item1")
    assert response["content"] == "Item1"
    api.sync()
    assert "Item1" in [i["content"] for i in api.state["items"]]
    item1 = [i for i in api.state["items"] if i["content"] == "Item1"][0]
    assert api.items.get_by_id(item1["id"]) == item1

    assert api.items.get(item1["id"])["item"]["content"] == item1["content"]

    item1.delete()
    api.commit()


def test_item_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.sync()

    item1.delete()
    response = api.commit()
    assert response["items"][0]["id"] == item1["id"]
    assert response["items"][0]["is_deleted"] == 1
    assert "Item1" not in [i["content"] for i in api.state["items"]]


def test_item_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()

    item1.update(content="UpdatedItem1")
    response = api.commit()
    assert response["items"][0]["content"] == "UpdatedItem1"
    assert "UpdatedItem1" in [i["content"] for i in api.state["items"]]
    assert api.items.get_by_id(item1["id"]) == item1

    item1.delete()
    api.commit()


def test_item_complete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    item2 = api.items.add("Item2", parent_id=item1["id"])
    api.commit()

    item2.complete()
    response = api.commit()
    assert response["items"][0]["content"] == "Item2"
    assert response["items"][0]["checked"] == 1
    assert 1 in [i["checked"] for i in api.state["items"] if i["id"] == item2["id"]]

    item1.delete()
    api.commit()
    item2.delete()
    api.commit()


def test_item_uncomplete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    item2 = api.items.add("Item2", parent_id=item1["id"])
    api.commit()
    item2.complete()
    api.commit()

    item2.uncomplete()
    response = api.commit()
    assert response["items"][0]["content"] == "Item2"
    assert response["items"][0]["checked"] == 0
    assert 0 in [i["checked"] for i in api.state["items"] if i["id"] == item1["id"]]

    item1.delete()
    api.commit()
    item2.delete()
    api.commit()


def test_item_archive(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    item2 = api.items.add("Item2", parent_id=item1["id"])
    api.commit()
    item2.complete()
    api.commit()

    item2.archive()
    response = api.commit()
    assert response["items"][0]["content"] == "Item2"
    assert response["items"][0]["in_history"] == 1
    assert 1 in [i["in_history"] for i in api.state["items"] if i["id"] == item2["id"]]

    item1.delete()
    api.commit()
    item2.delete()
    api.commit()


def test_item_unarchive(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    item2 = api.items.add("Item2", parent_id=item1["id"])
    api.commit()
    item2.complete()
    api.commit()
    item2.archive()
    api.commit()

    item2.unarchive()
    response = api.commit()
    assert response["items"][0]["content"] == "Item2"
    assert response["items"][0]["in_history"] == 0
    assert 0 in [i["in_history"] for i in api.state["items"] if i["id"] == item2["id"]]

    item1.delete()
    api.commit()
    item2.delete()
    api.commit()


def test_item_move_to_project(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    project1 = api.projects.add("Project1")
    api.commit()

    item1.move(project_id=project1["id"])
    response = api.commit()
    assert response["items"][0]["content"] == "Item1"
    assert response["items"][0]["project_id"] == project1["id"]
    assert project1["id"] in [
        i["project_id"] for i in api.state["items"] if i["id"] == item1["id"]
    ]

    item1.delete()
    api.commit()
    project1.delete()
    api.commit()


def test_item_move_to_section(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    section1 = api.sections.add("Section1", api.state["user"]["inbox_project"])
    api.commit()

    item1.move(section_id=section1["id"])
    response = api.commit()
    assert response["items"][0]["content"] == "Item1"
    assert response["items"][0]["section_id"] == section1["id"]
    assert section1["id"] in [
        i["section_id"] for i in api.state["items"] if i["id"] == item1["id"]
    ]

    item1.delete()
    api.commit()
    section1.delete()
    api.commit()


def test_item_move_to_parent(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    item2 = api.items.add("Item2")
    api.commit()

    item2.move(parent_id=item1["id"])
    response = api.commit()
    assert response["items"][0]["content"] == "Item2"
    assert response["items"][0]["parent_id"] == item1["id"]
    assert item1["id"] in [
        i["parent_id"] for i in api.state["items"] if i["id"] == item2["id"]
    ]

    item1.delete()
    api.commit()
    item2.delete()
    api.commit()


def test_item_update_date_complete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1", due={"string": "every day"})
    api.commit()

    now = time.time()
    tomorrow = time.gmtime(now + 24 * 3600)
    new_date_utc = time.strftime("%Y-%m-%dT%H:%M:%SZ", tomorrow)
    due = {
        "date": new_date_utc,
        "string": "every day",
    }
    api.items.update_date_complete(item1["id"], due=due)
    response = api.commit()
    assert response["items"][0]["due"]["string"] == "every day"
    assert "every day" in [
        i["due"]["string"] for i in api.state["items"] if i["id"] == item1["id"]
    ]

    item1.delete()
    api.commit()


def test_item_reorder(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    item2 = api.items.add("Item2")
    api.commit()

    api.items.reorder(
        items=[
            {"id": item1["id"], "child_order": 2},
            {"id": item2["id"], "child_order": 1},
        ]
    )
    response = api.commit()
    for item in response["items"]:
        if item["id"] == item1["id"]:
            assert item["child_order"] == 2
        if item["id"] == item2["id"]:
            assert item["child_order"] == 1
    assert 2 in [p["child_order"] for p in api.state["items"] if p["id"] == item1["id"]]
    assert 1 in [p["child_order"] for p in api.state["items"] if p["id"] == item2["id"]]

    item1.delete()
    api.commit()
    item2.delete()
    api.commit()


def test_item_update_day_orders(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    item2 = api.items.add("Item2")
    api.commit()

    api.items.update_day_orders({item1["id"]: 1, item2["id"]: 2})
    response = api.commit()
    for item in response["items"]:
        if item["id"] == item1["id"]:
            assert item["day_order"] == 1
        if item["id"] == item2["id"]:
            assert item["day_order"] == 2
    assert 1 == api.state["day_orders"][str(item1["id"])]
    assert 2 == api.state["day_orders"][str(item2["id"])]

    item1.delete()
    api.commit()

    item2.delete()
    api.commit()


def test_label_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    label1 = api.labels.add("Label1")
    response = api.commit()
    assert response["labels"][0]["name"] == "Label1"
    assert "Label1" in [l["name"] for l in api.state["labels"]]
    assert api.labels.get_by_id(label1["id"]) == label1

    assert api.labels.get(label1["id"])["label"]["name"] == label1["name"]

    label1.delete()
    api.commit()


def test_label_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    label1 = api.labels.add("Label1")
    api.commit()

    label1.delete()
    response = api.commit()
    assert response["labels"][0]["id"] == label1["id"]
    assert response["labels"][0]["is_deleted"] == 1
    assert "UpdatedLabel1" not in [l["name"] for l in api.state["labels"]]


def test_label_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    label1 = api.labels.add("Label1")
    api.commit()

    label1.update(name="UpdatedLabel1")
    response = api.commit()
    assert response["labels"][0]["name"] == "UpdatedLabel1"
    assert "UpdatedLabel1" in [l["name"] for l in api.state["labels"]]
    assert api.labels.get_by_id(label1["id"]) == label1

    label1.delete()
    api.commit()


def test_label_update_orders(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    label1 = api.labels.add("Label1")
    api.commit()
    label2 = api.labels.add("Label2")
    api.commit()

    api.labels.update_orders({label1["id"]: 1, label2["id"]: 2})
    response = api.commit()
    for label in response["labels"]:
        if label["id"] == label1["id"]:
            assert label["item_order"] == 1
        if label["id"] == label2["id"]:
            assert label["item_order"] == 2
    assert 1 in [
        l["item_order"] for l in api.state["labels"] if l["id"] == label1["id"]
    ]
    assert 2 in [
        l["item_order"] for l in api.state["labels"] if l["id"] == label2["id"]
    ]

    label1.delete()
    api.commit()
    label2.delete()
    api.commit()


def test_note_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()

    note1 = api.notes.add(item1["id"], "Note1")
    response = api.commit()
    assert response["notes"][0]["content"] == "Note1"
    assert "Note1" in [n["content"] for n in api.state["notes"]]
    assert api.notes.get_by_id(note1["id"]) == note1

    assert api.notes.get(note1["id"])["note"]["content"] == note1["content"]

    note1.delete()
    api.commit()
    item1.delete()
    api.commit()


def test_note_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    note1 = api.notes.add(item1["id"], "Note1")
    api.commit()

    note1.delete()
    response = api.commit()
    assert response["notes"][0]["id"] == note1["id"]
    assert response["notes"][0]["is_deleted"] == 1
    assert "UpdatedNote1" not in [n["content"] for n in api.state["notes"]]

    note1.delete()
    api.commit()
    item1.delete()
    api.commit()


def test_note_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1")
    api.commit()
    note1 = api.notes.add(item1["id"], "Note1")
    api.commit()

    note1.update(content="UpdatedNote1")
    response = api.commit()
    assert response["notes"][0]["content"] == "UpdatedNote1"
    assert "UpdatedNote1" in [n["content"] for n in api.state["notes"]]
    assert api.notes.get_by_id(note1["id"]) == note1

    note1.delete()
    api.commit()
    item1.delete()
    api.commit()


def test_projectnote_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    note1 = api.project_notes.add(project1["id"], "Note1")
    response = api.commit()
    assert response["project_notes"][0]["content"] == "Note1"
    assert "Note1" in [n["content"] for n in api.state["project_notes"]]
    assert api.project_notes.get_by_id(note1["id"]) == note1

    assert api.project_notes.get(note1["id"])["note"]["content"] == note1["content"]

    note1.delete()
    api.commit()
    project1.delete()
    api.commit()


def test_projectnote_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()
    note1 = api.project_notes.add(project1["id"], "Note1")
    api.commit()

    note1.delete()
    response = api.commit()
    assert response["project_notes"][0]["id"] == note1["id"]
    assert response["project_notes"][0]["is_deleted"] == 1
    assert "UpdatedNote1" not in [n["content"] for n in api.state["project_notes"]]

    project1.delete()
    api.commit()


def test_projectnote_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    project1 = api.projects.add("Project1")
    api.commit()
    note1 = api.project_notes.add(project1["id"], "Note1")
    api.commit()

    note1.update(content="UpdatedNote1")
    response = api.commit()
    assert response["project_notes"][0]["content"] == "UpdatedNote1"
    assert "UpdatedNote1" in [n["content"] for n in api.state["project_notes"]]
    assert api.project_notes.get_by_id(note1["id"]) == note1

    note1.delete()
    api.commit()
    project1.delete()
    api.commit()


def test_section_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    response = api.sections.add("Section1", api.state["user"]["inbox_project"])
    assert response["name"] == "Section1"
    api.commit()
    assert "Section1" in [i["name"] for i in api.state["sections"]]
    section1 = [i for i in api.state["sections"] if i["name"] == "Section1"][0]
    assert api.sections.get_by_id(section1["id"]) == section1

    assert api.sections.get(section1["id"])["section"]["name"] == section1["name"]

    section1.delete()
    api.commit()


def test_section_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    section1 = api.sections.add("Section1", api.state["user"]["inbox_project"])
    api.commit()

    section1.delete()
    response = api.commit()
    assert response["sections"][0]["id"] == section1["id"]
    assert response["sections"][0]["is_deleted"] == 1
    assert "Section1" not in [i["name"] for i in api.state["sections"]]


def test_section_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    section1 = api.sections.add("Section1", api.state["user"]["inbox_project"])
    api.commit()

    section1.update(name="UpdatedSection1")
    response = api.commit()
    assert response["sections"][0]["name"] == "UpdatedSection1"
    assert "UpdatedSection1" in [i["name"] for i in api.state["sections"]]
    assert api.sections.get_by_id(section1["id"]) == section1

    section1.delete()
    api.commit()


def test_section_archive(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    section1 = api.sections.add("Section1", api.state["user"]["inbox_project"])
    api.commit()

    section1.archive()
    response = api.commit()
    assert response["sections"][0]["name"] == "Section1"
    assert response["sections"][0]["is_archived"] == 1

    section1.delete()
    api.commit()


def test_section_unarchive(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    section1 = api.sections.add("Section1", api.state["user"]["inbox_project"])
    api.commit()

    section1.unarchive()
    response = api.commit()
    assert response["sections"][0]["name"] == "Section1"
    assert response["sections"][0]["is_archived"] == 0

    section1.delete()
    api.commit()


def test_section_move_to_project(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    section1 = api.sections.add("Section1", api.state["user"]["inbox_project"])
    api.commit()
    project1 = api.projects.add("Project1")
    api.commit()

    section1.move(project_id=project1["id"])
    response = api.commit()
    assert response["sections"][0]["name"] == "Section1"
    assert response["sections"][0]["project_id"] == project1["id"]
    assert project1["id"] in [
        i["project_id"] for i in api.state["sections"] if i["id"] == section1["id"]
    ]

    section1.delete()
    api.commit()
    project1.delete()
    api.commit()


def test_section_reorder(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    section1 = api.sections.add("Section1", api.state["user"]["inbox_project"])
    api.commit()
    section2 = api.sections.add("Section2", api.state["user"]["inbox_project"])
    api.commit()

    api.sections.reorder(
        sections=[
            {"id": section1["id"], "section_order": 2},
            {"id": section2["id"], "section_order": 1},
        ]
    )
    response = api.commit()
    for section in response["sections"]:
        if section["id"] == section1["id"]:
            assert section["section_order"] == 2
        if section["id"] == section2["id"]:
            assert section["section_order"] == 1
    assert 2 in [
        p["section_order"] for p in api.state["sections"] if p["id"] == section1["id"]
    ]
    assert 1 in [
        p["section_order"] for p in api.state["sections"] if p["id"] == section2["id"]
    ]

    section1.delete()
    api.commit()
    section2.delete()
    api.commit()


def test_filter_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    filter1 = api.filters.add("Filter1", "no due date")
    response = api.commit()
    assert response["filters"][0]["name"] == "Filter1"
    assert "Filter1" in [f["name"] for f in api.state["filters"]]
    assert api.filters.get_by_id(filter1["id"]) == filter1
    assert api.filters.get(filter1["id"])["filter"]["name"] == filter1["name"]

    filter1.delete()
    api.commit()


def test_filter_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    filter1 = api.filters.add("Filter1", "no due date")
    api.commit()

    filter1.delete()
    response = api.commit()
    assert response["filters"][0]["id"] == filter1["id"]
    assert response["filters"][0]["is_deleted"] == 1
    assert "Filter1" not in [p["name"] for p in api.state["filters"]]


def test_filter_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    filter1 = api.filters.add("Filter1", "no due date")
    api.commit()

    filter1.update(name="UpdatedFilter1")
    response = api.commit()
    assert response["filters"][0]["name"] == "UpdatedFilter1"
    assert "UpdatedFilter1" in [f["name"] for f in api.state["filters"]]
    assert api.filters.get_by_id(filter1["id"]) == filter1

    filter1.delete()
    api.commit()


def test_filter_update_orders(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    filter1 = api.filters.add("Filter1", "no due date")
    api.commit()

    filter2 = api.filters.add("Filter2", "today")
    api.commit()

    api.filters.update_orders({filter1["id"]: 2, filter2["id"]: 1})
    response = api.commit()
    for filter in response["filters"]:
        if filter["id"] == filter1["id"]:
            assert filter["item_order"] == 2
        if filter["id"] == filter2["id"]:
            assert filter["item_order"] == 1
    assert 2 in [
        f["item_order"] for f in api.state["filters"] if f["id"] == filter1["id"]
    ]
    assert 1 in [
        f["item_order"] for f in api.state["filters"] if f["id"] == filter2["id"]
    ]

    filter1.delete()
    api.commit()
    filter2.delete()
    api.commit()


def test_reminder_relative_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1", due={"string": "tomorrow 5pm"})
    api.commit()

    reminder1 = api.reminders.add(item1["id"], minute_offset=30)
    response = api.commit()
    assert response["reminders"][0]["minute_offset"] == 30
    assert reminder1["id"] in [p["id"] for p in api.state["reminders"]]
    assert api.reminders.get_by_id(reminder1["id"]) == reminder1
    assert api.reminders.get(reminder1["id"])["reminder"]["due"] == reminder1["due"]

    reminder1.delete()
    api.commit()
    item1.delete()
    api.commit()


def test_reminder_relative_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1", due={"string": "tomorrow 5pm"})
    api.commit()
    reminder1 = api.reminders.add(item1["id"], minute_offset=30)
    api.commit()

    reminder1.delete()
    response = api.commit()
    assert response["reminders"][0]["is_deleted"] == 1
    assert reminder1["id"] not in [p["id"] for p in api.state["reminders"]]

    item1.delete()
    api.commit()


def test_reminder_relative_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1", due={"string": "tomorrow 5pm"})
    api.commit()
    reminder1 = api.reminders.add(item1["id"], minute_offset=30)
    api.commit()

    reminder1.update(minute_offset=str(15))
    response = api.commit()
    assert response["reminders"][0]["minute_offset"] == 15
    assert reminder1["id"] in [p["id"] for p in api.state["reminders"]]
    assert api.reminders.get_by_id(reminder1["id"]) == reminder1

    reminder1.delete()
    api.commit()
    item1.delete()
    api.commit()


def test_reminder_absolute_add(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1", due={"string": "tomorrow 5pm"})
    api.commit()

    now = time.time()
    tomorrow = time.gmtime(now + 24 * 3600)
    due_date_utc = time.strftime("%Y-%m-%dT%H:%M:%SZ", tomorrow)
    reminder1 = api.reminders.add(item1["id"], due={"date": due_date_utc})
    response = api.commit()
    assert response["reminders"][0]["due"]["date"] == due_date_utc
    tomorrow = time.gmtime(time.time() + 24 * 3600)
    assert reminder1["id"] in [p["id"] for p in api.state["reminders"]]
    assert api.reminders.get_by_id(reminder1["id"]) == reminder1
    assert api.reminders.get(reminder1["id"])["reminder"]["due"] == reminder1["due"]

    reminder1.delete()
    api.commit()
    item1.delete()
    api.commit()


def test_reminder_absolute_delete(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1", due={"string": "tomorrow 5pm"})
    api.commit()

    now = time.time()
    tomorrow = time.gmtime(now + 24 * 3600)
    due_date_utc = time.strftime("%Y-%m-%dT%H:%M:%SZ", tomorrow)
    reminder1 = api.reminders.add(item1["id"], due={"date": due_date_utc})
    api.commit()

    api.reminders.delete(reminder1["id"])
    response = api.commit()
    assert response["reminders"][0]["is_deleted"] == 1
    assert reminder1["id"] not in [p["id"] for p in api.state["reminders"]]

    item1.delete()
    response = api.commit()


def test_reminder_absolute_update(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api.sync()

    item1 = api.items.add("Item1", due={"string": "tomorrow 5pm"})
    api.commit()

    now = time.time()
    tomorrow = time.gmtime(now + 24 * 3600)
    due_date_utc = time.strftime("%Y-%m-%dT%H:%M:%SZ", tomorrow)
    reminder1 = api.reminders.add(item1["id"], due={"date": due_date_utc})
    api.commit()

    tomorrow = time.gmtime(now + 24 * 3600 + 60)
    due_date_utc = time.strftime("%Y-%m-%dT%H:%M:%SZ", tomorrow)
    api.reminders.update(reminder1["id"], due_date_utc=due_date_utc)
    response = api.commit()
    assert response["reminders"][0]["due"]["date"] == due_date_utc
    assert reminder1["id"] in [p["id"] for p in api.state["reminders"]]
    assert api.reminders.get_by_id(reminder1["id"]) == reminder1

    reminder1.delete()
    api.commit()
    item1.delete()
    api.commit()


def test_locations(api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)

    api.sync()

    api.locations.clear()
    api.commit()

    assert api.state["locations"] == []


def test_live_notifications(api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)

    api.sync()

    api.live_notifications.set_last_read(api.state["live_notifications_last_read_id"])
    response = api.commit()
    assert (
        response["live_notifications_last_read_id"]
        == api.state["live_notifications_last_read_id"]
    )


def test_share_accept(cleanup, cleanup2, api_endpoint, api_token, api_token2):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api2 = todoist.api.TodoistAPI(api_token2, api_endpoint)

    api.user.update(auto_invite_disabled=1)
    api.commit()
    api.sync()

    api2.user.update(auto_invite_disabled=1)
    api2.commit()
    api2.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    api.projects.share(project1["id"], api2.state["user"]["email"])
    response = api.commit()
    assert response["projects"][0]["name"] == project1["name"]
    assert response["projects"][0]["shared"]

    response2 = api2.sync()
    invitation1 = next(
        (
            ln
            for ln in response2["live_notifications"]
            if ln["notification_type"] == "share_invitation_sent"
        ),
        None,
    )
    assert invitation1 is not None
    assert invitation1["project_name"] == project1["name"]
    assert invitation1["from_user"]["email"] == api.state["user"]["email"]

    api2.invitations.accept(invitation1["id"], invitation1["invitation_secret"])
    response2 = api2.commit()
    assert api2.state["user"]["id"] in [
        p["user_id"] for p in api2.state["collaborator_states"]
    ]

    api.sync()
    project1 = [p for p in api.state["projects"] if p["name"] == "Project1"][0]
    project1.delete()
    api.commit()


def test_share_reject(cleanup, cleanup2, api_endpoint, api_token, api_token2):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api2 = todoist.api.TodoistAPI(api_token2, api_endpoint)

    api.user.update(auto_invite_disabled=1)
    api.commit()
    api.sync()

    api2.user.update(auto_invite_disabled=1)
    api2.commit()
    api2.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    api.projects.share(project1["id"], api2.state["user"]["email"])
    response = api.commit()
    assert response["projects"][0]["name"] == project1["name"]
    assert response["projects"][0]["shared"]

    response2 = api2.sync()
    invitation2 = next(
        (
            ln
            for ln in response2["live_notifications"]
            if ln["notification_type"] == "share_invitation_sent"
        ),
        None,
    )
    assert invitation2 is not None
    assert invitation2["project_name"] == project1["name"]
    assert invitation2["from_user"]["email"] == api.state["user"]["email"]

    api2.invitations.reject(invitation2["id"], invitation2["invitation_secret"])
    response2 = api2.commit()
    assert len(response2["projects"]) == 0
    assert len(response2["collaborator_states"]) == 0

    project1 = [p for p in api.state["projects"] if p["name"] == "Project1"][0]
    project1.delete()
    api.commit()


def test_share_delete(cleanup, cleanup2, api_endpoint, api_token, api_token2):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)
    api2 = todoist.api.TodoistAPI(api_token2, api_endpoint)

    api.user.update(auto_invite_disabled=1)
    api.commit()
    api.sync()

    api2.user.update(auto_invite_disabled=1)
    api2.commit()
    api2.sync()

    project1 = api.projects.add("Project1")
    api.commit()

    api.projects.share(project1["id"], api2.state["user"]["email"])
    response = api.commit()
    assert response["projects"][0]["name"] == project1["name"]
    assert response["projects"][0]["shared"]

    response2 = api2.sync()
    invitation3 = next(
        (
            ln
            for ln in response2["live_notifications"]
            if ln["notification_type"] == "share_invitation_sent"
        ),
        None,
    )
    assert invitation3 is not None
    assert invitation3["project_name"] == project1["name"]
    assert invitation3["from_user"]["email"] == api.state["user"]["email"]

    api.invitations.delete(invitation3["id"])
    api.commit()

    project1 = [p for p in api.state["projects"] if p["name"] == "Project1"][0]
    project1.delete()
    api.commit()


def test_items_archive(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)

    # Create and complete five tasks
    project = api.projects.add("Project")
    items = [
        api.items.add("task{}".format(i), project_id=project["id"]) for i in range(5)
    ]
    for i, item in enumerate(items):
        date_completed = "2019-01-01T00:00:0{}Z".format(i)
        api.items.complete(item_id=item["id"], date_completed=date_completed)
    api.commit()

    # Create an archive manager to iterate over them
    manager = api.items_archive.for_project(project["id"])
    item_ids = [item["id"] for item in manager.items()]
    assert item_ids == [item["id"] for item in items[::-1]]

    # tear down
    project.delete()
    api.commit()


def test_sections_archive(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)

    # Create and complete five sections
    project = api.projects.add("Project")
    sections = [
        api.sections.add("s{}".format(i), project_id=project["id"]) for i in range(5)
    ]
    for i, section in enumerate(sections):
        date_archived = "2019-01-01T00:00:0{}Z".format(i)
        api.sections.archive(section_id=section["id"], date_archived=date_archived)
    api.commit()

    # Create an archive manager to iterate over them
    manager = api.sections_archive.for_project(project["id"])
    section_ids = [section["id"] for section in manager.sections()]
    assert section_ids == [section["id"] for section in sections[::-1]]

    # tear down
    project.delete()
    api.commit()


def test_templates(cleanup, api_endpoint, api_token):
    api = todoist.api.TodoistAPI(api_token, api_endpoint)

    api.sync()

    project1 = api.projects.add("Project1")
    project2 = api.projects.add("Project2")
    api.commit()

    item1 = api.items.add("Item1", project_id=project1["id"])
    api.commit()

    template = api.templates.export_as_file(project1["id"])
    assert "task,Item1,4,1" in template
    with io.open("/tmp/example.csv", "w", encoding="utf-8") as example:
        example.write(template)

    result = api.templates.import_into_project(project1["id"], "/tmp/example.csv")
    assert result == {"status": u"ok"}

    item1.delete()
    api.commit()
    project1.delete()
    api.commit()
    project2.delete()
    api.commit()



# -*- coding: utf-8 -*-
import json

from .generic import Manager


class BusinessUsersManager(Manager):
    def invite(self, email_list):
        """
        Send a business user invitation.
        """
        params = {"token": self.token, "email_list": json.dumps(email_list)}
        return self.api._get("business/users/invite", params=params)

    def accept_invitation(self, id, secret):
        """
        Accept a business user invitation.
        """
        params = {"token": self.token, "id": id, "secret": secret}
        return self.api._get("business/users/accept_invitation", params=params)

    def reject_invitation(self, id, secret):
        """
        Reject a business user invitation.
        """
        params = {"token": self.token, "id": id, "secret": secret}
        return self.api._get("business/users/reject_invitation", params=params)



# -*- coding: utf-8 -*-
from .generic import Manager


class UserManager(Manager):
    def update(self, **kwargs):
        """
        Updates the user data.
        """
        cmd = {
            "type": "user_update",
            "uuid": self.api.generate_uuid(),
            "args": kwargs,
        }
        self.queue.append(cmd)

    def update_goals(self, **kwargs):
        """
        Updates the user's karma goals.
        """
        cmd = {
            "type": "update_goals",
            "uuid": self.api.generate_uuid(),
            "args": kwargs,
        }
        self.queue.append(cmd)

    def sync(self):
        return self.api.sync()

    def get(self, key=None, default=None):
        ret = self.state["user"]
        if key is not None:
            ret = ret.get(key, default)
        return ret

    def get_id(self):
        return self.state["user"]["id"]

    def login(self, email, password):
        """
        Logins user, and returns the response received by the server.
        """
        data = self.api._post("user/login", data={"email": email, "password": password})
        if "token" in data:
            self.api.token = data["token"]
        return data

    def login_with_google(self, email, oauth2_token, **kwargs):
        """
        Logins user with Google account, and returns the response received by
        the server.

        """
        data = {"email": email, "oauth2_token": oauth2_token}
        data.update(kwargs)
        data = self.api._post("user/login_with_google", data=data)
        if "token" in data:
            self.api.token = data["token"]
        return data

    def register(self, email, full_name, password, **kwargs):
        """
        Registers a new user.
        """
        data = {"email": email, "full_name": full_name, "password": password}
        data.update(kwargs)
        data = self.api._post("user/register", data=data)
        if "token" in data:
            self.api.token = data["token"]
        return data

    def delete(self, current_password, **kwargs):
        """
        Deletes an existing user.
        """
        params = {"token": self.token, "current_password": current_password}
        params.update(kwargs)
        return self.api._get("user/delete", params=params)

    def update_notification_setting(self, notification_type, service, dont_notify):
        """
        Updates the user's notification settings.
        """
        return self.api._post(
            "user/update_notification_setting",
            data={
                "token": self.token,
                "notification_type": notification_type,
                "service": service,
                "dont_notify": dont_notify,
            },
        )



# -*- coding: utf-8 -*-
from .generic import Manager


class EmailsManager(Manager):
    def get_or_create(self, obj_type, obj_id, **kwargs):
        """
        Get or create email to an object.
        """
        params = {"token": self.token, "obj_type": obj_type, "obj_id": obj_id}
        params.update(kwargs)
        return self.api._get("emails/get_or_create", params=params)

    def disable(self, obj_type, obj_id, **kwargs):
        """
        Disable email to an object.
        """
        params = {"token": self.token, "obj_type": obj_type, "obj_id": obj_id}
        params.update(kwargs)
        return self.api._get("emails/disable", params=params)



# -*- coding: utf-8 -*-
from .. import models
from .generic import AllMixin, GetByIdMixin, Manager, SyncMixin


class ProjectsManager(Manager, AllMixin, GetByIdMixin, SyncMixin):

    state_name = "projects"
    object_type = "project"

    def add(self, name, **kwargs):
        """
        Creates a local project object.
        """
        obj = models.Project({"name": name}, self.api)
        obj.temp_id = obj["id"] = "$" + self.api.generate_uuid()
        obj.data.update(kwargs)
        self.state[self.state_name].append(obj)
        cmd = {
            "type": "project_add",
            "temp_id": obj.temp_id,
            "uuid": self.api.generate_uuid(),
            "args": {key: obj.data[key] for key in obj.data if key != "id"},
        }
        self.queue.append(cmd)
        return obj

    def update(self, project_id, **kwargs):
        """
        Updates a project remotely.
        """
        obj = self.get_by_id(project_id)
        if obj:
            obj.data.update(kwargs)

        args = {"id": project_id}
        args.update(kwargs)
        cmd = {
            "type": "project_update",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def delete(self, project_id):
        """
        Deletes a project remotely.
        """
        cmd = {
            "type": "project_delete",
            "uuid": self.api.generate_uuid(),
            "args": {"id": project_id},
        }
        self.queue.append(cmd)

    def archive(self, project_id):
        """
        Marks project as archived remotely.
        """
        cmd = {
            "type": "project_archive",
            "uuid": self.api.generate_uuid(),
            "args": {"id": project_id},
        }
        self.queue.append(cmd)

    def unarchive(self, project_id):
        """
        Marks project as unarchived remotely.
        """
        cmd = {
            "type": "project_unarchive",
            "uuid": self.api.generate_uuid(),
            "args": {"id": project_id},
        }
        self.queue.append(cmd)

    def move(self, project_id, parent_id):
        """
        Moves project to another parent.
        """
        args = {
            "id": project_id,
            "parent_id": parent_id,
        }
        cmd = {"type": "project_move", "uuid": self.api.generate_uuid(), "args": args}
        self.queue.append(cmd)

    def reorder(self, projects):
        """
        Updates the child_order of the specified projects.
        """
        cmd = {
            "type": "project_reorder",
            "uuid": self.api.generate_uuid(),
            "args": {"projects": projects},
        }
        self.queue.append(cmd)

    def share(self, project_id, email):
        """
        Shares a project with a user.
        """
        cmd = {
            "type": "share_project",
            "temp_id": self.api.generate_uuid(),
            "uuid": self.api.generate_uuid(),
            "args": {"project_id": project_id, "email": email},
        }
        self.queue.append(cmd)

    def get_archived(self):
        """
        Returns archived projects.
        """
        params = {"token": self.token}
        return self.api._get("projects/get_archived", params=params)

    def get_data(self, project_id):
        """
        Returns a project's uncompleted items.
        """
        params = {"token": self.token, "project_id": project_id}
        return self.api._get("projects/get_data", params=params)

    def get(self, project_id):
        """
        Gets an existing project.
        """
        params = {"token": self.token, "project_id": project_id}
        obj = self.api._get("projects/get", params=params)
        if obj and "error" in obj:
            return None
        data = {"projects": [], "project_notes": []}
        if obj.get("project"):
            data["projects"].append(obj.get("project"))
        if obj.get("notes"):
            data["project_notes"] += obj.get("notes")
        self.api._update_state(data)
        return obj



"""
Managers to get the list of archived items and sections.

Manager makers available as "items_archive" and "sections_archive" attributes of
API object.


Usage example (for items).

```python

# Create an API object
import todoist
api = todoist.TodoistAPI(...)

# Get project ID (take inbox)
project_id = api.user.get()['inbox_project']

# Initiate ItemsArchiveManager
archive = api.items_archive.for_project(project_id)

# Iterate over the list of completed items for the archive
for item in archive.items():
    print(item["date_completed"], item["content"])
```
"""
from typing import TYPE_CHECKING, Dict, Iterator, Optional

from ..models import Item, Model, Section

if TYPE_CHECKING:
    from ..api import TodoistAPI


class ArchiveManager(object):

    object_model = Model

    def __init__(self, api, element_type):
        # type: (TodoistAPI, str) -> None
        assert element_type in {"sections", "items"}
        self.api = api
        self.element_type = element_type

    def next_page(self, cursor):
        # type: (Optional[str]) -> Dict
        """Return response for the next page of the archive."""
        resp = self.api.session.get(
            self._next_url(),
            params=self._next_query_params(cursor),
            headers=self._request_headers(),
        )
        resp.raise_for_status()
        return resp.json()

    def _next_url(self):
        return "{0}/sync/{1}/archive/{2}".format(
            self.api.api_endpoint, self.api.api_version, self.element_type
        )

    def _next_query_params(self, cursor):
        # type: (Optional[str]) -> Dict
        ret = {}
        if cursor:
            ret["cursor"] = cursor
        return ret

    def _request_headers(self):
        return {"Authorization": "Bearer {}".format(self.api.token)}

    def _iterate(self):
        has_more = True
        cursor = None

        while True:
            if not has_more:
                break

            resp = self.next_page(cursor)

            elements = [self._make_element(data) for data in resp[self.element_type]]
            has_more = resp["has_more"]
            cursor = resp.get("next_cursor")
            for el in elements:
                yield el

    def _make_element(self, data):
        return self.object_model(data, self.api)


class SectionsArchiveManagerMaker(object):
    def __init__(self, api):
        self.api = api

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)

    def for_project(self, project_id):
        """Get manager to iterate over all archived sections for project."""
        return SectionsArchiveManager(api=self.api, project_id=project_id)


class SectionsArchiveManager(ArchiveManager):

    object_model = Section

    def __init__(self, api, project_id):
        super(SectionsArchiveManager, self).__init__(api, "sections")
        self.project_id = project_id

    def __repr__(self):
        return "SectionsArchiveManager(project_id={})".format(self.project_id)

    def sections(self):
        # type: () -> Iterator[Section]
        """Iterate over all archived sections."""
        for obj in self._iterate():
            yield obj

    def _next_query_params(self, cursor):
        ret = super(SectionsArchiveManager, self)._next_query_params(cursor)
        ret["project_id"] = self.project_id
        return ret


class ItemsArchiveManagerMaker(object):
    def __init__(self, api):
        self.api = api

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)

    def for_project(self, project_id):
        """Get manager to iterate over all top-level archived items for project."""
        return ItemsArchiveManager(api=self.api, project_id=project_id)

    def for_section(self, section_id):
        """Get manager to iterate over all top-level archived items for section."""
        return ItemsArchiveManager(api=self.api, section_id=section_id)

    def for_parent(self, parent_id):
        """Get manager to iterate over all archived sub-tasks for an item."""
        return ItemsArchiveManager(api=self.api, parent_id=parent_id)


class ItemsArchiveManager(ArchiveManager):

    object_model = Item

    def __init__(self, api, project_id=None, section_id=None, parent_id=None):
        super(ItemsArchiveManager, self).__init__(api, "items")
        assert sum([bool(project_id), bool(section_id), bool(parent_id)]) == 1
        self.project_id = project_id
        self.section_id = section_id
        self.parent_id = parent_id

    def __repr__(self):
        k, v = self._key_value()
        return "ItemsArchiveManager({}={})".format(k, v)

    def items(self):
        # type: () -> Iterator[Item]
        """Iterate over all archived items."""
        for obj in self._iterate():
            yield obj

    def _next_query_params(self, cursor):
        ret = super(ItemsArchiveManager, self)._next_query_params(cursor)
        k, v = self._key_value()
        ret[k] = v
        return ret

    def _key_value(self):
        if self.project_id:
            return "project_id", self.project_id
        elif self.section_id:
            return "section_id", self.section_id
        else:  # if self.parent_id:
            return "parent_id", self.parent_id



# -*- coding: utf-8 -*-
from .. import models
from .generic import AllMixin, GetByIdMixin, Manager, SyncMixin


class GenericNotesManager(Manager, AllMixin, GetByIdMixin, SyncMixin):

    object_type = "note"

    def update(self, note_id, **kwargs):
        """
        Updates an note remotely.
        """
        args = {"id": note_id}
        args.update(kwargs)
        cmd = {
            "type": "note_update",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def delete(self, note_id):
        """
        Deletes an note remotely.
        """
        cmd = {
            "type": "note_delete",
            "uuid": self.api.generate_uuid(),
            "args": {"id": note_id},
        }
        self.queue.append(cmd)


class NotesManager(GenericNotesManager):

    state_name = "notes"

    def add(self, item_id, content, **kwargs):
        """
        Creates a local item note object.
        """
        obj = models.Note({"item_id": item_id, "content": content}, self.api)
        obj.temp_id = obj["id"] = self.api.generate_uuid()
        obj.data.update(kwargs)
        self.state[self.state_name].append(obj)
        cmd = {
            "type": "note_add",
            "temp_id": obj.temp_id,
            "uuid": self.api.generate_uuid(),
            "args": {key: obj.data[key] for key in obj.data if key != "id"},
        }
        self.queue.append(cmd)
        return obj

    def get(self, note_id):
        """
        Gets an existing note.
        """
        params = {"token": self.token, "note_id": note_id}
        obj = self.api._get("notes/get", params=params)
        if obj and "error" in obj:
            return None
        data = {"notes": []}
        if obj.get("note"):
            data["notes"].append(obj.get("note"))
        self.api._update_state(data)
        return obj


class ProjectNotesManager(GenericNotesManager):

    state_name = "project_notes"

    def add(self, project_id, content, **kwargs):
        """
        Creates a local project note object.
        """
        obj = models.ProjectNote(
            {"project_id": project_id, "content": content}, self.api
        )
        obj.temp_id = obj["id"] = self.api.generate_uuid()
        obj.data.update(kwargs)
        self.state[self.state_name].append(obj)
        cmd = {
            "type": "note_add",
            "temp_id": obj.temp_id,
            "uuid": self.api.generate_uuid(),
            "args": {key: obj.data[key] for key in obj.data if key != "id"},
        }
        self.queue.append(cmd)
        return obj

    def get(self, note_id):
        """
        Gets an existing project note.
        """
        params = {"token": self.token, "note_id": note_id}
        obj = self.api._get("notes/get", params=params)
        if obj and "error" in obj:
            return None
        data = {"project_notes": []}
        if obj.get("note"):
            data["project_notes"].append(obj.get("note"))
        self.api._update_state(data)
        return obj



import datetime
import functools
import json
import os
import uuid

import requests

from todoist import models
from todoist.managers.activity import ActivityManager
from todoist.managers.archive import (
    ItemsArchiveManagerMaker,
    SectionsArchiveManagerMaker,
)
from todoist.managers.backups import BackupsManager
from todoist.managers.biz_invitations import BizInvitationsManager
from todoist.managers.business_users import BusinessUsersManager
from todoist.managers.collaborator_states import CollaboratorStatesManager
from todoist.managers.collaborators import CollaboratorsManager
from todoist.managers.completed import CompletedManager
from todoist.managers.emails import EmailsManager
from todoist.managers.filters import FiltersManager
from todoist.managers.invitations import InvitationsManager
from todoist.managers.items import ItemsManager
from todoist.managers.labels import LabelsManager
from todoist.managers.live_notifications import LiveNotificationsManager
from todoist.managers.locations import LocationsManager
from todoist.managers.notes import NotesManager, ProjectNotesManager
from todoist.managers.projects import ProjectsManager
from todoist.managers.quick import QuickManager
from todoist.managers.reminders import RemindersManager
from todoist.managers.sections import SectionsManager
from todoist.managers.templates import TemplatesManager
from todoist.managers.uploads import UploadsManager
from todoist.managers.user import UserManager
from todoist.managers.user_settings import UserSettingsManager

DEFAULT_API_VERSION = "v8"


class SyncError(Exception):
    pass


class TodoistAPI(object):
    """
    Implements the API that makes it possible to interact with a Todoist user
    account and its data.
    """

    _serialize_fields = ("token", "api_endpoint", "sync_token", "state", "temp_ids")

    @classmethod
    def deserialize(cls, data):
        obj = cls()
        for key in cls._serialize_fields:
            if key in data:
                setattr(obj, key, data[key])
        return obj

    def __init__(
        self,
        token="",
        api_endpoint="https://api.todoist.com",
        api_version=DEFAULT_API_VERSION,
        session=None,
        cache="~/.todoist-sync/",
    ):
        self.api_endpoint = api_endpoint
        self.api_version = api_version
        self.reset_state()
        self.token = token  # User's API token
        self.temp_ids = {}  # Mapping of temporary ids to real ids
        self.queue = []  # Requests to be sent are appended here
        self.session = session or requests.Session()  # Session instance for requests

        # managers
        self.biz_invitations = BizInvitationsManager(self)
        self.collaborators = CollaboratorsManager(self)
        self.collaborator_states = CollaboratorStatesManager(self)
        self.filters = FiltersManager(self)
        self.invitations = InvitationsManager(self)
        self.items = ItemsManager(self)
        self.labels = LabelsManager(self)
        self.live_notifications = LiveNotificationsManager(self)
        self.locations = LocationsManager(self)
        self.notes = NotesManager(self)
        self.projects = ProjectsManager(self)
        self.project_notes = ProjectNotesManager(self)
        self.reminders = RemindersManager(self)
        self.sections = SectionsManager(self)
        self.user = UserManager(self)
        self.user_settings = UserSettingsManager(self)

        self.activity = ActivityManager(self)
        self.backups = BackupsManager(self)
        self.business_users = BusinessUsersManager(self)
        self.completed = CompletedManager(self)
        self.emails = EmailsManager(self)
        self.quick = QuickManager(self)
        self.templates = TemplatesManager(self)
        self.uploads = UploadsManager(self)

        self.items_archive = ItemsArchiveManagerMaker(self)
        self.sections_archive = SectionsArchiveManagerMaker(self)

        if cache:  # Read and write user state on local disk cache
            self.cache = os.path.expanduser(cache)
            self._read_cache()
        else:
            self.cache = None

    def reset_state(self):
        self.sync_token = "*"
        self.state = {  # Local copy of all of the user's objects
            "collaborator_states": [],
            "collaborators": [],
            "day_orders": {},
            "day_orders_timestamp": "",
            "filters": [],
            "items": [],
            "labels": [],
            "live_notifications": [],
            "live_notifications_last_read_id": -1,
            "locations": [],
            "notes": [],
            "project_notes": [],
            "projects": [],
            "reminders": [],
            "sections": [],
            "settings_notifications": {},
            "user": {},
            "user_settings": {},
        }

    def __getitem__(self, key):
        return self.state[key]

    def serialize(self):
        return {key: getattr(self, key) for key in self._serialize_fields}

    def get_api_url(self):
        return "{0}/sync/{1}/".format(self.api_endpoint, self.api_version)

    def _update_state(self, syncdata):
        """
        Updates the local state, with the data returned by the server after a
        sync.
        """
        # Check sync token first
        if "sync_token" in syncdata:
            self.sync_token = syncdata["sync_token"]

        # It is straightforward to update these type of data, since it is
        # enough to just see if they are present in the sync data, and then
        # either replace the local values or update them.
        if "day_orders" in syncdata:
            self.state["day_orders"].update(syncdata["day_orders"])
        if "day_orders_timestamp" in syncdata:
            self.state["day_orders_timestamp"] = syncdata["day_orders_timestamp"]
        if "live_notifications_last_read_id" in syncdata:
            self.state["live_notifications_last_read_id"] = syncdata[
                "live_notifications_last_read_id"
            ]
        if "locations" in syncdata:
            self.state["locations"] = syncdata["locations"]
        if "settings_notifications" in syncdata:
            self.state["settings_notifications"].update(
                syncdata["settings_notifications"]
            )
        if "user" in syncdata:
            self.state["user"].update(syncdata["user"])
        if "user_settings" in syncdata:
            self.state["user_settings"].update(syncdata["user_settings"])

        # Updating these type of data is a bit more complicated, since it is
        # necessary to find out whether an object in the sync data is new,
        # updates an existing object, or marks an object to be deleted.  But
        # the same procedure takes place for each of these types of data.
        resp_models_mapping = [
            ("collaborators", models.Collaborator),
            ("collaborator_states", models.CollaboratorState),
            ("filters", models.Filter),
            ("items", models.Item),
            ("labels", models.Label),
            ("live_notifications", models.LiveNotification),
            ("notes", models.Note),
            ("project_notes", models.ProjectNote),
            ("projects", models.Project),
            ("reminders", models.Reminder),
            ("sections", models.Section),
        ]
        for datatype, model in resp_models_mapping:
            if datatype not in syncdata:
                continue

            # Process each object of this specific type in the sync data.
            for remoteobj in syncdata[datatype]:
                # Find out whether the object already exists in the local
                # state.
                localobj = self._find_object(datatype, remoteobj)
                if localobj is not None:
                    # If the object is already present in the local state, then
                    # we either update it, or if marked as to be deleted, we
                    # remove it.
                    is_deleted = remoteobj.get("is_deleted", 0)
                    if is_deleted == 0 or is_deleted is False:
                        localobj.data.update(remoteobj)
                    else:
                        self.state[datatype].remove(localobj)
                else:
                    # If not, then the object is new and it should be added,
                    # unless it is marked as to be deleted (in which case it's
                    # ignored).
                    is_deleted = remoteobj.get("is_deleted", 0)
                    if is_deleted == 0 or is_deleted is False:
                        newobj = model(remoteobj, self)
                        self.state[datatype].append(newobj)

    def _read_cache(self):
        if not self.cache:
            return

        try:
            os.makedirs(self.cache)
        except OSError:
            if not os.path.isdir(self.cache):
                raise

        try:
            with open(self.cache + self.token + ".json") as f:
                state = f.read()
            state = json.loads(state)
            self._update_state(state)

            with open(self.cache + self.token + ".sync") as f:
                sync_token = f.read()
            self.sync_token = sync_token
        except Exception:
            return

    def _write_cache(self):
        if not self.cache:
            return
        result = json.dumps(self.state, indent=2, sort_keys=True, default=state_default)
        with open(self.cache + self.token + ".json", "w") as f:
            f.write(result)
        with open(self.cache + self.token + ".sync", "w") as f:
            f.write(self.sync_token)

    def _find_object(self, objtype, obj):
        """
        Searches for an object in the local state, depending on the type of
        object, and then on its primary key is.  If the object is found it is
        returned, and if not, then None is returned.
        """
        if objtype == "collaborators":
            return self.collaborators.get_by_id(obj["id"])
        elif objtype == "collaborator_states":
            return self.collaborator_states.get_by_ids(
                obj["project_id"], obj["user_id"]
            )
        elif objtype == "filters":
            return self.filters.get_by_id(obj["id"], only_local=True)
        elif objtype == "items":
            return self.items.get_by_id(obj["id"], only_local=True)
        elif objtype == "labels":
            return self.labels.get_by_id(obj["id"], only_local=True)
        elif objtype == "live_notifications":
            return self.live_notifications.get_by_id(obj["id"])
        elif objtype == "notes":
            return self.notes.get_by_id(obj["id"], only_local=True)
        elif objtype == "project_notes":
            return self.project_notes.get_by_id(obj["id"], only_local=True)
        elif objtype == "projects":
            return self.projects.get_by_id(obj["id"], only_local=True)
        elif objtype == "reminders":
            return self.reminders.get_by_id(obj["id"], only_local=True)
        elif objtype == "sections":
            return self.sections.get_by_id(obj["id"], only_local=True)
        else:
            return None

    def _replace_temp_id(self, temp_id, new_id):
        """
        Replaces the temporary id generated locally when an object was first
        created, with a real Id supplied by the server.  True is returned if
        the temporary id was found and replaced, and False otherwise.
        """
        # Go through all the objects for which we expect the temporary id to be
        # replaced by a real one.
        for datatype in [
            "filters",
            "items",
            "labels",
            "notes",
            "project_notes",
            "projects",
            "reminders",
            "sections",
        ]:
            for obj in self.state[datatype]:
                if obj.temp_id == temp_id:
                    obj["id"] = new_id
                    return True
        return False

    def _get(self, call, url=None, **kwargs):
        """
        Sends an HTTP GET request to the specified URL, and returns the JSON
        object received (if any), or whatever answer it got otherwise.
        """
        if not url:
            url = self.get_api_url()

        response = self.session.get(url + call, **kwargs)

        try:
            return response.json()
        except ValueError:
            return response.text

    def _post(self, call, url=None, **kwargs):
        """
        Sends an HTTP POST request to the specified URL, and returns the JSON
        object received (if any), or whatever answer it got otherwise.
        """
        if not url:
            url = self.get_api_url()

        response = self.session.post(url + call, **kwargs)

        try:
            return response.json()
        except ValueError:
            return response.text

    # Sync
    def generate_uuid(self):
        """
        Generates a uuid.
        """
        return str(uuid.uuid1())

    def sync(self, commands=None):
        """
        Sends to the server the changes that were made locally, and also
        fetches the latest updated data from the server.
        """
        post_data = {
            "token": self.token,
            "sync_token": self.sync_token,
            "day_orders_timestamp": self.state["day_orders_timestamp"],
            "include_notification_settings": 1,
            "resource_types": json_dumps(["all"]),
            "commands": json_dumps(commands or []),
        }
        response = self._post("sync", data=post_data)
        if "temp_id_mapping" in response:
            for temp_id, new_id in response["temp_id_mapping"].items():
                self.temp_ids[temp_id] = new_id
                self._replace_temp_id(temp_id, new_id)
        self._update_state(response)
        self._write_cache()
        return response

    def commit(self, raise_on_error=True):
        """
        Commits all requests that are queued.  Note that, without calling this
        method none of the changes that are made to the objects are actually
        synchronized to the server, unless one of the aforementioned Sync API
        calls are called directly.
        """
        if len(self.queue) == 0:
            return
        ret = self.sync(commands=self.queue)
        del self.queue[:]
        if "sync_status" in ret:
            if raise_on_error:
                for k, v in ret["sync_status"].items():
                    if v != "ok":
                        raise SyncError(k, v)
        return ret

    # Miscellaneous

    def query(self, queries, **kwargs):
        """
        DEPRECATED: query endpoint is deprecated for a long time and this
        method will be removed in the next major version of todoist-python
        """
        params = {"queries": json_dumps(queries), "token": self.token}
        params.update(kwargs)
        return self._get("query", params=params)

    def add_item(self, content, **kwargs):
        """
        Adds a new task.
        """
        params = {"token": self.token, "content": content}
        params.update(kwargs)
        if "labels" in params:
            params["labels"] = str(params["labels"])
        return self._get("add_item", params=params)

    # Class
    def __repr__(self):
        name = self.__class__.__name__
        unsaved = "*" if len(self.queue) > 0 else ""
        email = self.user.get("email")
        email_repr = repr(email) if email else "<not synchronized>"
        return "%s%s(%s)" % (name, unsaved, email_repr)


def state_default(obj):
    return obj.data


def json_default(obj):
    if isinstance(obj, datetime.datetime):
        return obj.strftime("%Y-%m-%dT%H:%M:%S")
    elif isinstance(obj, datetime.date):
        return obj.strftime("%Y-%m-%d")
    elif isinstance(obj, datetime.time):
        return obj.strftime("%H:%M:%S")


json_dumps = functools.partial(json.dumps, separators=",:", default=json_default)



# -*- coding: utf-8 -*-
from .generic import AllMixin, GetByIdMixin, Manager, SyncMixin


class LiveNotificationsManager(Manager, GetByIdMixin, AllMixin, SyncMixin):

    state_name = "live_notifications"
    object_type = None  # there is no object type associated

    def set_last_read(self, id):
        """
        Sets the last known notification.
        """
        cmd = {
            "type": "live_notifications_set_last_read",
            "uuid": self.api.generate_uuid(),
            "args": {"id": id},
        }
        self.queue.append(cmd)

    def mark_read(self, id):
        """
        Marks notification as read.
        """
        cmd = {
            "type": "live_notifications_mark_read",
            "uuid": self.api.generate_uuid(),
            "args": {"id": id},
        }
        self.queue.append(cmd)

    def mark_read_all(self):
        """
        Marks all notifications as read.
        """
        cmd = {
            "type": "live_notifications_mark_read_all",
            "uuid": self.api.generate_uuid(),
        }
        self.queue.append(cmd)

    def mark_unread(self, id):
        """
        Marks notification as unread.
        """
        cmd = {
            "type": "live_notifications_mark_unread",
            "uuid": self.api.generate_uuid(),
            "args": {"id": id},
        }
        self.queue.append(cmd)



# -*- coding: utf-8 -*-
from .. import models
from .generic import AllMixin, GetByIdMixin, Manager, SyncMixin


class LabelsManager(Manager, AllMixin, GetByIdMixin, SyncMixin):

    state_name = "labels"
    object_type = "label"

    def add(self, name, **kwargs):
        """
        Creates a local label object.
        """
        obj = models.Label({"name": name}, self.api)
        obj.temp_id = obj["id"] = self.api.generate_uuid()
        obj.data.update(kwargs)
        self.state[self.state_name].append(obj)
        cmd = {
            "type": "label_add",
            "temp_id": obj.temp_id,
            "uuid": self.api.generate_uuid(),
            "args": {key: obj.data[key] for key in obj.data if key != "id"},
        }
        self.queue.append(cmd)
        return obj

    def update(self, label_id, **kwargs):
        """
        Updates a label remotely.
        """
        args = {"id": label_id}
        args.update(kwargs)
        cmd = {
            "type": "label_update",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def delete(self, label_id):
        """
        Deletes a label remotely.
        """
        cmd = {
            "type": "label_delete",
            "uuid": self.api.generate_uuid(),
            "args": {"id": label_id},
        }
        self.queue.append(cmd)

    def update_orders(self, id_order_mapping):
        """
        Updates the orders of multiple labels remotely.
        """
        cmd = {
            "type": "label_update_orders",
            "uuid": self.api.generate_uuid(),
            "args": {"id_order_mapping": id_order_mapping},
        }
        self.queue.append(cmd)

    def get(self, label_id):
        """
        Gets an existing label.
        """
        params = {"token": self.token, "label_id": label_id}
        obj = self.api._get("labels/get", params=params)
        if obj and "error" in obj:
            return None
        data = {"labels": []}
        if obj.get("label"):
            data["labels"].append(obj.get("label"))
        self.api._update_state(data)
        return obj



# -*- coding: utf-8 -*-
from .generic import Manager, SyncMixin


class CollaboratorStatesManager(Manager, SyncMixin):

    state_name = "collaborator_states"
    object_type = None  # there is no object type associated

    def get_by_ids(self, project_id, user_id):
        """
        Finds and returns the collaborator state based on the project and user
        ids.
        """
        for obj in self.state[self.state_name]:
            if obj["project_id"] == project_id and obj["user_id"] == user_id:
                return obj
        return None



# -*- coding: utf-8 -*-
from .generic import Manager


class TemplatesManager(Manager):
    def import_into_project(self, project_id, filename, **kwargs):
        """
        Imports a template into a project.
        """
        data = {"token": self.token, "project_id": project_id}
        data.update(kwargs)
        files = {"file": open(filename, "r")}
        return self.api._post("templates/import_into_project", data=data, files=files)

    def export_as_file(self, project_id, **kwargs):
        """
        Exports a template as a file.
        """
        data = {"token": self.token, "project_id": project_id}
        data.update(kwargs)
        return self.api._post("templates/export_as_file", data=data)

    def export_as_url(self, project_id, **kwargs):
        """
        Exports a template as a URL.
        """
        data = {"token": self.token, "project_id": project_id}
        data.update(kwargs)
        return self.api._post("templates/export_as_url", data=data)



# -*- coding: utf-8 -*-
import os

from setuptools import setup


def read(fname):
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except Exception:
        return ""


setup(
    name="todoist-python",
    version="8.1.4",
    packages=["todoist", "todoist.managers"],
    author="Doist Team",
    author_email="integrations@todoist.com",
    license="MIT",
    description="todoist-python - The official Todoist Python API library",
    long_description=read("README.md"),
    install_requires=[
        "requests",
        "typing;python_version<'3.5'",
    ],
    # see here for complete list of classifiers
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=(
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
    ),
)



# -*- coding: utf-8 -*-
from .. import models
from .generic import AllMixin, GetByIdMixin, Manager, SyncMixin


class FiltersManager(Manager, AllMixin, GetByIdMixin, SyncMixin):

    state_name = "filters"
    object_type = "filter"

    def add(self, name, query, **kwargs):
        """
        Creates a local filter object.
        """
        obj = models.Filter({"name": name, "query": query}, self.api)
        obj.temp_id = obj["id"] = self.api.generate_uuid()
        obj.data.update(kwargs)
        self.state[self.state_name].append(obj)
        cmd = {
            "type": "filter_add",
            "temp_id": obj.temp_id,
            "uuid": self.api.generate_uuid(),
            "args": {key: obj.data[key] for key in obj.data if key != "id"},
        }
        self.queue.append(cmd)
        return obj

    def update(self, filter_id, **kwargs):
        """
        Updates a filter remotely.
        """
        args = {"id": filter_id}
        args.update(kwargs)
        cmd = {
            "type": "filter_update",
            "uuid": self.api.generate_uuid(),
            "args": args,
        }
        self.queue.append(cmd)

    def delete(self, filter_id):
        """
        Deletes a filter remotely.
        """
        cmd = {
            "type": "filter_delete",
            "uuid": self.api.generate_uuid(),
            "args": {"id": filter_id},
        }
        self.queue.append(cmd)

    def update_orders(self, id_order_mapping):
        """
        Updates the orders of multiple filters remotely.
        """
        cmd = {
            "type": "filter_update_orders",
            "uuid": self.api.generate_uuid(),
            "args": {"id_order_mapping": id_order_mapping},
        }
        self.queue.append(cmd)

    def get(self, filter_id):
        """
        Gets an existing filter.
        """
        params = {"token": self.token, "filter_id": filter_id}
        obj = self.api._get("filters/get", params=params)
        if obj and "error" in obj:
            return None
        data = {"filters": []}
        if obj.get("filter"):
            data["filters"].append(obj.get("filter"))
        self.api._update_state(data)
        return obj



# -*- coding: utf-8 -*-
from .generic import Manager


class UserSettingsManager(Manager):
    def update(self, **kwargs):
        """
        Updates the user's settings.
        """
        cmd = {
            "type": "user_settings_update",
            "uuid": self.api.generate_uuid(),
            "args": kwargs,
        }
        self.queue.append(cmd)



# -*- coding: utf-8 -*-
from .generic import Manager


class UploadsManager(Manager):
    def add(self, filename, **kwargs):
        """
        Uploads a file.

        param filename: (str) name of file to upload
        """
        data = {"token": self.token}
        data.update(kwargs)
        files = {"file": open(filename, "rb")}
        return self.api._post("uploads/add", data=data, files=files)

    def get(self, **kwargs):
        """
        Returns all user's uploads.

        kwargs:
            limit: (int, optional) number of results (1-50)
            last_id: (int, optional) return results with id<last_id
        """
        params = {"token": self.token}
        params.update(kwargs)
        return self.api._get("uploads/get", params=params)

    def delete(self, file_url):
        """
        Deletes upload.

        param file_url: (str) uploaded file URL
        """
        params = {"token": self.token, "file_url": file_url}
        return self.api._get("uploads/delete", params=params)
